package cz.muni.fi.pb162.exceptions;


/**
 * The test class MyPostOfficeTest.
 *
 * @author  (your name)
 * @version (a version number or a date)
 */
public class MyPostOfficeTest extends junit.framework.TestCase
{
    public MyPostOfficeTest(String testName) {
        super(testName);
    }
    
    public void testRules() {
        BasicRulesTester.testMethodsAndAttributes(MailException.class);
        BasicRulesTester.testMethodsAndAttributes(BadAddressException.class);
        BasicRulesTester.testMethodsAndAttributes(RecipientFailureException.class);
        BasicRulesTester.testMethodsAndAttributes(LetterRefusedException.class);
        BasicRulesTester.testMethodsAndAttributes(RecipientMissedException.class);
        BasicRulesTester.testMethodsAndAttributes(MyPostOffice.class);
    }
    
    private void assertAncestor(Class ancestor, Class clazz) {
        assertEquals("Trida " + clazz + " neni primym potomkem tridy " + ancestor,
		ancestor,clazz.getSuperclass());
    }
    
    public void testExceptionHierarchy() {
        assertAncestor(Exception.class, MailException.class);
        assertAncestor(MailException.class,BadAddressException.class);
        assertAncestor(MailException.class,RecipientFailureException.class);
        assertAncestor(MailException.class,RecipientMissedException.class);
        assertAncestor(MailException.class,LetterRefusedException.class);
    }

    public void testNewMyPostOfficeWithNullAddressBook() {
        try {
            PostOffice post = new MyPostOffice(null, 1);
            fail("Podarilo se vytvorit instanci tridy MyPostOffice s parametrem " +
		    "addressBook = null, aniz by konstruktor vyhodil vyjimku " +
		    "NullPointerException");
		  } catch (NullPointerException e) {
		      assertEquals("Vyjimka NullPointerException vyhozena konstruktorem " +
		      "MyPostOffice(AddressBook,int) by jako popis chyby mela " +
		      "obsahovat nazev chybneho parametru","addressBook",e.getMessage());
		      // All is ok
		  }
    }
    
    public void testNewMyPostOfficeWithWrongRetries() {
        try {
            PostOffice post = new MyPostOffice(new SimpleAddressBook(), -1);
            fail("Podarilo se vytvorit instanci tridy MyPostOffice se zapornym" +
            " poctem opakovani (retries), aniz by konstruktor vyhodil " +
            "vyjimku IllegalArgumentException");
        } catch (IllegalArgumentException e) {
            assertEquals("Vyjimka IllegalArgumentException vyhozena konstruktorem " +
            "MyPostOffice(AddressBook,int) by jako popis chyby mela " +
            "obsahovat nazev chybneho parametru",
            "maxSendAttemptCount",e.getMessage());
            // All is ok
        }
        try {
            PostOffice post = new MyPostOffice(new SimpleAddressBook(), 0);
            fail("Podarilo se vytvorit instanci tridy MyPostOffice s nulovym" +
            " poctem opakovani (retries), aniz by konstruktor vyhodil " +
            "vyjimku IllegalArgumentException");
        } catch (IllegalArgumentException e) {
            assertEquals("Vyjimka IllegalArgumentException vyhozena konstruktorem " +
            "MyPostOffice(AddressBook,int) by jako popis chyby mela " +
            "obsahovat nazev chybneho parametru",
            "maxSendAttemptCount",e.getMessage());
            // All is ok
        }
    }
    
    public void testDeliverLetter() {
        TestABook aBook = new TestABook();
        PostOffice post = new MyPostOffice(aBook, 1);
        RegisteredLetter letter = new RegisteredLetter("messageText","address1");
        
        try {
            post.deliver(letter);
        } catch (Exception ex) {
            fail("Pri pokusu odeslat dopis byla neocekavane vyhozena vyjimka " + ex);
        }
        assertEquals("Pri odeslani dopisu nebyl dorucovateli predan spravny dopis",
                     letter,aBook.deliveredLetter);
        assertEquals("Kontrolovana adresa neodpovida adresatovi z dopisu",
                     "address1",aBook.address);
        assertEquals("Metoda AddressBook.checkAddress() byla zbytecne volana vicekrat",
                     1,aBook.getConnectionCounter);
        assertEquals("Metoda Mailman.deliverLetter() byla zbytecne volana vicekrat",
                     1,aBook.deliveryCounter);
    }

    public void testDeliveryWithBadAddressException() {
        TestABook aBook = new TestABook();
        PostOffice post = new MyPostOffice(aBook, 5);
        RegisteredLetter letter = new RegisteredLetter("messageText","address1");
        
        aBook.connectionException = new BadAddressException("");
        
        try {
            post.deliver(letter);
            fail("Metoda MyPostOffice.deliver() nevyhodi vyjimku, " +
            "ackoliv adresa prijemmce neexistuje a metoda " +
            "AddressBook.checkAddress() vyhodila vyjimku BadAddressException");
        } catch (BadAddressException ex) {
		    assertNotNull("Vyjimka BadAddressException " +
		    "nema nastavenou smysluplnou zpravu o chybe",ex.getMessage());
		} catch (Exception ex) {
		    fail("Metoda MyPostOffice.deliver() misto ocekavane vyjimky " +
		    "BadAddressException vyhodila vyjimku " + ex);
	    }
	    assertEquals("Metoda Mailman.deliverLetter() byla zbytecne volana vicekrat",
	    1,aBook.getConnectionCounter);
    }
    
    public void testDeliveryWithLetterRefusedException() {
        TestABook aBook = new TestABook();
        PostOffice post = new MyPostOffice(aBook, 5);
        RegisteredLetter letter = new RegisteredLetter("messageText","address1");
        
        aBook.deliveryException = new LetterRefusedException("");
        
        try {
            post.deliver(letter);
            fail("Metoda MyPostOffice.deliver() nevyhodi vyjimku, " +
            "ackoliv prijemce odmita dopis prevzit a metoda " +
            "Mailman.deliverLetter() vyhodila vyjimku LetterRefusedException");
        } catch (RecipientFailureException ex) {
            assertSame("Vyjimka RecipientFailureException " +
		    "nema spravne nastavenou svoji pricinu",
		    aBook.deliveryException, ex.getCause());
		    assertNotNull("Vyjimka RecipientFailureException " +
		    "nema nastavenou smysluplnou zpravu o chybe",ex.getMessage());
		} catch (Exception ex) {
		    fail("Metoda MyPostOffice.deliver() misto ocekavane vyjimky " +
		    "RecipientFailureException vyhodila vyjimku " + ex);
	    }
	    assertEquals("Metoda Mailman.deliverLetter() byla zbytecne volana vicekrat",
	    1,aBook.deliveryCounter);
    }
    
    public void testFailedDeliveryAndNoRetry() {
        TestABook aBook = new TestABook();
        PostOffice post = new MyPostOffice(aBook, 1);
        RegisteredLetter letter = new RegisteredLetter("messageText","address1");
        
        aBook.deliveryException = new RecipientMissedException("");
        aBook.deliveryExceptionCounter = 1;
        
        try {
            post.deliver(letter);
            fail("Metoda MyPostOffice.deliver() nevyhodi vyjimku, " +
            "ackoliv se nepodarilo zastihnout adresata a pocet " +
            "pokusu byl nastaven na jedna");
        } catch (RecipientFailureException ex) {
            assertSame("Vyjimka RecipientFailureException " +
		    "nema spravne nastavenou svoji pricinu",
		    aBook.deliveryException, ex.getCause());
		    assertNotNull("Vyjimka RecipientFailureException " +
		    "nema nastavenou smysluplnou zpravu o chybe",ex.getMessage());
		} catch (Exception ex) {
		    fail("Metoda MyPostOffice.deliver() misto ocekavane vyjimky " +
		    "RecipientFailureException vyhodila vyjimku " + ex);
	    }
        assertEquals("Metoda AddressBook.checkAddress() byla zbytecne volana vicekrat",
        1,aBook.getConnectionCounter);
	    assertEquals("Metoda Mailman.deliverLetter() byla zbytecne volana vicekrat",
	    1,aBook.deliveryCounter);
    }

    public void testFailedDeliveryAndRetryIsOk() {
        TestABook aBook = new TestABook();
        PostOffice post = new MyPostOffice(aBook, 2);
        RegisteredLetter letter = new RegisteredLetter("messageText","address1");
        
        aBook.deliveryException = new RecipientMissedException("");
        aBook.deliveryExceptionCounter = 1;
        
        try {
            post.deliver(letter);
		} catch (Exception ex) {
		    fail("Pri pokusu dorucit dopis byla neocekavane vyhozena vyjimka " 
		    + ex + " (jde o situaci, kdy napoprve metoda " +
		    "Mailman.deliverLetter() vyhodi vyjimku RecipientMissedException " +
		    "a az teprve napodruhe uspeje)");
	    }
        assertEquals("Pri odeslani dopisu nebyl dorucovateli predan spravny dopis",
                     letter,aBook.deliveredLetter);
        assertEquals("Kontrolovana adresa neodpovida adresatovi z dopisu",
                     "address1",aBook.address);
        assertEquals("Metoda AddressBook.checkAddress() byla zbytecne volana vicekrat",
                     1,aBook.getConnectionCounter);
        assertEquals("Metoda Mailman.deliverLetter() byla zbytecne volana vicekrat",
                     2,aBook.deliveryCounter);
    }

    public void testFailedDeliveryAndRetryIsNotOk() {
        TestABook aBook = new TestABook();
        PostOffice post = new MyPostOffice(aBook, 2);
        RegisteredLetter letter = new RegisteredLetter("messageText","address1");
        
        aBook.deliveryException = new RecipientMissedException("");
        aBook.deliveryExceptionCounter = 2;
        
        try {
            post.deliver(letter);
            fail("Metoda MyPostOffice.deliver() nevyhodi vyjimku, " +
            "ackoliv se nepodarilo zastihnout adresata ani napodruhe a pocet " +
            "pokusu byl nastaven na dve");
        } catch (RecipientFailureException ex) {
            assertSame("Vyjimka RecipientFailureException " +
		    "nema spravne nastavenou svoji pricinu",
		    aBook.deliveryException, ex.getCause());
		    assertNotNull("Vyjimka RecipientFailureException " +
		    "nema nastavenou smysluplnou zpravu o chybe",ex.getMessage());
		} catch (Exception ex) {
		    fail("Metoda MyPostOffice.deliver() misto ocekavane vyjimky " +
		    "RecipientFailureException vyhodila vyjimku " + ex);
	    }
        assertEquals("Metoda AddressBook.checkAddress() byla zbytecne volana vicekrat",
        1,aBook.getConnectionCounter);
	    assertEquals("Metoda Mailman.deliverLetter() byla zbytecne volana vicekrat",
	    2,aBook.deliveryCounter);
    }

    
    
    
    private static class TestABook implements AddressBook {
        
        private RegisteredLetter deliveredLetter;
        private String address;
        private int getConnectionCounter;
        private int deliveryCounter;
        private MailException connectionException;
        private MailException deliveryException;
        private int deliveryExceptionCounter;

        public Mailman checkAddress(String recipient) throws BadAddressException {
            getConnectionCounter++;
            if (connectionException instanceof BadAddressException) {
                throw (BadAddressException) connectionException;
            } else {
                address = recipient;
                return new Mailman() {
                    public void deliverLetter(RegisteredLetter letter) 
                        throws RecipientMissedException, LetterRefusedException 
                    {
                        deliveryCounter++;
                        if (deliveryException instanceof LetterRefusedException) {
                            throw (LetterRefusedException) deliveryException;
                        } else if (deliveryException instanceof RecipientMissedException &&
                                   deliveryExceptionCounter-- > 0) {
                            throw (RecipientMissedException) deliveryException;
                        }                                   
                        deliveredLetter = letter;
                    }	       
                };
            }
		}
    }
}
